/*
 * NOSA HEADER START
 *
 * The contents of this file are subject to the terms of the NASA Open 
 * Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may 
 * not use this file except in compliance with the Agreement.
 *
 * You can obtain a copy of the agreement at
 *   docs/NASA_Open_Source_Agreement_1.3.txt
 * or 
 *   http://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
 *
 * See the Agreement for the specific language governing permissions
 * and limitations under the Agreement.
 *
 * When distributing Covered Code, include this NOSA HEADER in each
 * file and include the Agreement file at 
 * docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the 
 * following below this NOSA HEADER, with the fields enclosed by 
 * brackets "[]" replaced with your own identifying information: 
 * Portions Copyright [yyyy] [name of copyright owner]
 *
 * NOSA HEADER END
 *
 * Copyright (c) 2011-2012 United States Government as represented by 
 * the National Aeronautics and Space Administration. No copyright is 
 * claimed in the United States under Title 17, U.S.Code. All Other 
 * Rights Reserved.
 *
 * $Id: $
 */
package gsfc.spdf.edit.guis;

import javax.swing.JOptionPane;

import gsfc.nssdc.cdf.Attribute;
import gsfc.nssdc.cdf.CDF;
import gsfc.nssdc.cdf.CDFException;
import gsfc.nssdc.cdf.Entry;

import gsfc.spdf.istp.Idl;
import gsfc.spdf.istp.ISTPCompliance;
import gsfc.spdf.istp.Variable;
import gsfc.spdf.istp.VerificationCallback;


/**
 * SKTEditor's implementation of 
 * {@link gsfc.spdf.istp.VerificationCallback} for use during 
 * ISTP verification operations.
 *
 * @author B. Harris
 * @version $Revision: 1.1 $
 */
public class IstpVerificationCallback
    implements VerificationCallback {


    /**
     * Handles a variable name error condition.
     *
     * @param var the variable whose name failed the compliance test
     * @param conflictVar the variable whose name is in conflict with
     *            var.  null if name error is not caused by another
     *            variable.
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleVariableNameError(Variable var, 
                                       Variable conflictVar) {

        int reply = JOptionPane.showConfirmDialog(null,
                        "Variable name error: " +
                        var.getName() + ", " + 
                        conflictVar.getName(), 
                        "ISTP Compliance Error",
                        JOptionPane.YES_NO_CANCEL_OPTION);

        switch (reply) {

        case JOptionPane.YES_OPTION:

            // rename
            return CORRECTED;

        case JOptionPane.NO_OPTION:

            return NOT_CORRECTED;

        case JOptionPane.CANCEL_OPTION:

            return ABORT;
        }

        return NOT_CORRECTED;
    }


    /**
     * Handles a duplicate attribute name error condition.
     *
     * @param attr a conflicting attribute
     * @param conflictAttr the other conflicting attribute
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleDuplicateAttributeNameError(Attribute attr, 
                                       Attribute conflictAttr) {

        int reply = JOptionPane.showConfirmDialog(null,
                        "Duplicate Attribute name error: " +
                        attr.getName() + ", " +
                        conflictAttr.getName(), 
                        "ISTP Compliance Error",
                        JOptionPane.YES_NO_CANCEL_OPTION);

        switch (reply) {

        case JOptionPane.YES_OPTION:

            // rename
            return CORRECTED;

        case JOptionPane.NO_OPTION:

            return NOT_CORRECTED;

        case JOptionPane.CANCEL_OPTION:

            return ABORT;
        }

        return NOT_CORRECTED;
    }


    /**
     * Handles an attribute name error condition.
     *
     * @param attr the affending attribute
     * @param cause description of problem
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleAttributeNameError(Attribute attr, 
                                        String cause) {

        String name = attr.getName();  // attribute name

        do {

            name = (String)JOptionPane.showInputDialog(null,
                        "Attribute '" + name + 
                        "' has an invalid name: " + cause +
                        "\nEnter a new name for the attribute", 
                        "ISTP Compliance Error",
                        JOptionPane.QUESTION_MESSAGE);

            if (name == null) {

                return ABORT;
            }
/*
            else if (user wants to delete attribute) {

                attr.delete();
                return CORRECTED;
            }
*/
        }
        while ((cause = Idl.validateIdentifier(name)) != null);

        try {

            attr.rename(name);
        }
        catch (CDFException e) {

            JOptionPane.showMessageDialog(null,
                "Unexpected error while renaming attribute: " +
                e.getMessage(),
                "Attribute Rename Error",
                JOptionPane.ERROR_MESSAGE);

            return NOT_CORRECTED;
        }

        return CORRECTED;
    }


    /**
     * Handles a Variable Attribute value that contains non-ASCII 
     * characters.
     *
     * @param var the affending variable.
     * @param attr the affending attribute.
     * @param entry the affending entry.
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleNonAsciiAttributeValueError(
        Variable var, Attribute attr, Entry entry) {

        String varName = var.getName();// variable's name
        String attrName = attr.getName();
                                       // attribute's name
        String entryValue = null;      // entry value
        try {

            entryValue = (String)entry.getData();
        }
        catch (CDFException e) {

            // cannot correct it if unable to get it

            return CORRECTED;
        }
        do {

            entryValue = (String)
                JOptionPane.showInputDialog(null,
                    "Variable '" + varName + "'s '" +
                    attrName + "' value contains non-ASCII " +
                    " character(s)." +
                    "\nEnter a new value for the attribute", 
                    "ISTP Compliance Error",
                    JOptionPane.QUESTION_MESSAGE);

            if (entryValue == null) {

                return ABORT;
            }
        }
        while (!ISTPCompliance.isAscii(entryValue));

        try {

            entry.putData(CDF.CDF_CHAR, entryValue);
        }
        catch (CDFException e) {

            JOptionPane.showMessageDialog(null,
                "Unexpected error while changing attribute entry " +
                "value: " + e.getMessage(),
                "Variable Attribute Entry Error",
                JOptionPane.ERROR_MESSAGE);

            return NOT_CORRECTED;
        }

        return CORRECTED;
    }


    /**
     * Method that is called when a Variable is missing dimension size
     * information.
     *
     * @param var the affending variable.
     * @param dimensions the number of dimensions.
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleMissingDimSizeError(
        Variable var, long dimensions) {

        return NOT_CORRECTED;
    }


    /**
     * Method that is called when a Variable has a bad dimension size.
     *
     * @param var the affending variable.
     * @param dimension dimension having the bad size.
     * @param dimensionSize size of dimension.
     * @return CORRECTED, NOT_CORRECTED, or ABORT status to indicate
     *     the action taken and whether to proceed with the
     *     verification operation
     */
    public int handleBadDimSizeError(
        Variable var, int dimension, long dimensionSize) {

        return NOT_CORRECTED;
    }

}
