/*
 * NOSA HEADER START
 *
 * The contents of this file are subject to the terms of the NASA Open 
 * Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may 
 * not use this file except in compliance with the Agreement.
 *
 * You can obtain a copy of the agreement at
 *   docs/NASA_Open_Source_Agreement_1.3.txt
 * or 
 *   http://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
 *
 * See the Agreement for the specific language governing permissions
 * and limitations under the Agreement.
 *
 * When distributing Covered Code, include this NOSA HEADER in each
 * file and include the Agreement file at 
 * docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the 
 * following below this NOSA HEADER, with the fields enclosed by 
 * brackets "[]" replaced with your own identifying information: 
 * Portions Copyright [yyyy] [name of copyright owner]
 *
 * NOSA HEADER END
 *
 * Copyright (c) 2011-2016 United States Government as represented by 
 * the National Aeronautics and Space Administration. No copyright is 
 * claimed in the United States under Title 17, U.S.Code. All Other 
 * Rights Reserved.
 *
 * $Id: GetNotes.java,v 1.23 2016/12/14 14:02:29 bharris Exp $
 */
package gsfc.spdf.istp.tools;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import gsfc.nssdc.cdf.CDF;
import gsfc.nssdc.cdf.CDFConstants;
import gsfc.nssdc.cdf.CDFException;

import gsfc.spdf.istp.Filename;
import gsfc.spdf.util.TextUtils;


/**
 * This class represents an application that extracts certain global and
 * variable attribute information from the specified 
 * <a href="http://cdf.gsfc.nasa.gov/">Common Data Format (CDF)</a> 
 * files and produces an HTML representation of the information.
 */
public class GetNotes {

    /**
     * A map containing the {@link DatasetNotes} for each dataset.  The 
     * key to the map is the dataset's name.
     */
    private TreeMap datasetNotes = new TreeMap();

    /**
     * Flag indicating that the output should be XHTML instead of HTML.
     */
    private boolean outputXhtml = false;

    /**
     * Flag indicating that the output should be standalone HTML 
     * instead of fragmented HTML that depends upon include files.
     */
    private boolean outputHtml = false;


    /**
     * Constructs a GetNotes object.
     */
    public GetNotes() {}


    /**
     * Gets the value of the outputXhtml attribute.
     *
     * @return value of outputXhtml attribute
     * @see #setOutputXhtml
     */
    public boolean getOutputXhtml() {

        return outputXhtml;
    }


    /**
     * Sets the value of the outputXhtml attribute.
     *
     * @param value new outputXhtml value
     * @see #getOutputXhtml
     */
    public void setOutputXhtml(boolean value) {

        outputXhtml = value;
    }


    /**
     * Gets the value of the outputHtml attribute.
     *
     * @return value of outputHtml attribute
     * @see #setOutputHtml
     */
    public boolean getOutputHtml() {

        return outputHtml;
    }


    /**
     * Sets the value of the outputHtml attribute.
     *
     * @param value new outputHtml value
     * @see #getOutputHtml
     */
    public void setOutputHtml(boolean value) {

        outputHtml = value;
    }


    /**
     * Reads the notes (attribute values) from the specifed files and
     * stores them in the datasetNotes map.
     *
     * @param files name of CDF files to read notes from
     */
    public void readNotes(String[] files) {

        for (int i = 0; i < files.length; i++) {

            String filename = files[i];// the name of the file that is
                                       // currently being read

            String dataset = Filename.getDataset(filename);
                                       // dataset's name

            if (dataset == null) {

                System.err.println(
                    "Unrecognized filename pattern in file named '" +
                    filename + "'.");
                System.err.println("Skipping file.");
            }
            else {

                CDF cdf = null;        // CDF file

                try {

                    cdf = CDF.open(filename, CDFConstants.READONLYon);

                    DatasetNotes dNotes = null;
                                       // this dataset's notes

                    dNotes = (DatasetNotes)datasetNotes.get(dataset);

                    if (dNotes == null) {

                        dNotes = new DatasetNotes(dataset, cdf);

                        datasetNotes.put(dataset, dNotes);
                    }
                    else {

                        dNotes.mergeValue(cdf);
                    }
                }
                catch (CDFException e) {

                    System.err.println("Error accessing " + filename + ".");
                    System.err.println(e.getMessage());
                    System.err.println("Skipping file.");
                }
                finally {

                    if (cdf != null) {

                        try {

                            cdf.close();
                        }
                        catch (CDFException e) {

                            // at this point, I don't care
                        }
                    }
                }
            } // endif dataset != null
        } // endfor each file
    }


    /**
     * Writes an HTML representation of all the notes in the datasetNotes
     * map to a file with a name derived from the given filename.
     *
     * @param filename begining portion of output filename
     * @throws IOException if an IOException occurs writing the file
     */
    public void writeAllNotes(String filename) 
        throws IOException {

        writeNotes(datasetNotes, null, filename);
    }


    /**
     * Writes an HTML representation of the notes in the given 
     * DatasetNotes map to a file with a name derived from the given 
     * filename.
     *
     * @param dNotes map containing the DatasetNotes to write.
     * @param subsetId subset identifier.  null if not a subset.
     * @param filename begining portion of output filename.
     * @throws IOException if an IOException occurs writing the file
     */
    public void writeNotes(
        SortedMap dNotes, String subsetId, String filename) 
        throws IOException {

        PrintWriter out = null;        // PrintWriter used to output 
                                       // notes

        if (filename.equals("-")) {

            out = new PrintWriter(System.out);
        }
        else {

            String extension = outputXhtml ? ".xhtml" : ".html";   
                                       // filename's extension

            out = new PrintWriter(
                      new BufferedWriter(
                          new FileWriter(filename + extension))); 
        }

        String title = "CDAWeb Served Heliophysics Datasets";
                                       // title of page
        if (subsetId != null) {

            title = title + subsetId;
        }

        PrintHeader(out, title);

        out.println("<a name=\"top\"></a>");

        title = title.replaceAll("CDAWeb",
            "<a href=\"/\">CDAWeb</a>");

        PrintToc(dNotes, title, out);

        PrintNotes(dNotes, out);

        PrintFooter(out);

        out.close();
    }


    /**
     * Writes separate HTML pages for each subset of the notes in 
     * datasetNotes.  Each subsets is defined by the set of datasets
     * with names begining with each letter of the alphabet.  That is,
     * each dataset with a name beginning with 'A' belong to one subset
     * and therefore appear on one HTML page.  Each dataset with a name
     * beginning with 'B' belong to another subset and so on.  The HTML
     * page filename is constructed by concatenating the subset letter
     * to the end of the given filename value.
     *
     * @param filename begining portion of output filename
     * @throws IOException if an IOException occurs writing the file
     */
    public void writeAllSubNotes(String filename)
        throws IOException {

        for (char subset = 'A'; subset <= 'Z'; subset++) {

            SortedMap subDatasetNotes = 
                datasetNotes.subMap(subset + "", 
                                    (char)(subset + 1) + "");
                                       // subset of datasetNotes whose 
                                       // name begins with the subset 
                                       // character

            if (subDatasetNotes.size() > 0) {

                writeNotes(subDatasetNotes, 
                    " Beginning with '" + subset + "'", 
                    filename + subset);
            }
        } // endfor each subset
    }


    /**
     * Prints the header portion of an HTML page.
     *
     * @param out PrintWriter to use for printing
     * @param title title of HTML page
     */
    private void PrintHeader(PrintWriter out, String title) {

        if (outputXhtml || outputHtml) {

            String metaEnd = null;     // characters to end a <meta> 
                                       // element
            if (outputXhtml) {

                metaEnd = "/>";

                out.println(
                    "<?xml version=\"1.0\" encoding=\"utf-8\"?>");
                out.println("<!DOCTYPE html");
                out.println(
                    "PUBLIC \"-//W3C/DTD XHTML 1.0 Transitional//EN\"");
                out.println("\"http://www.w3.org/TR/xhtml1/DTD/" +
                        "xhtml1-transitional.dtd\">");
                out.println(
                    "<html xmlns=\"http://www.w3.org/1999/xhtml\" " +
                        "xml:lang=\"en\" lang=\"en\">");
            }
            else {

                metaEnd = ">";

                out.println("<!DOCTYPE html PUBLIC \"-//W3C//DTD " +
                    "HTML 4.01 Transitional//EN\" " +
                    "\"http://www.w3c.org/TR/1999/" +
                    "REC-html401-19991224/loose.dtd\">");
                out.println("<html lang=\"en\">");
            }

            out.println("<head>");
            out.println("  <title>" + title + "</title>");
            out.println("  <meta http-equiv=\"Content-Type\" " +
                "content=\"text/html\" " + metaEnd);
            out.println("  <meta name=\"orgcode\" content=\"672\" " + 
                metaEnd);
            out.println("  <meta name=\"description\" " +
                "content=\"Information about CDAWeb Heliophysics datasets.\" " +
                metaEnd);
            out.println(
                "  <meta name=\"keywords\" content=\"science, data, " +
                "CDAWeb, coordinated data analysis, multi-mission, " +
                "multi-instrument, space physics\" " + metaEnd);
            out.println(
                "  <meta name=\"rno\" content=\"Robert.E.McGuire.1\" " +
                metaEnd);
            out.println("  <meta name=\"content-owner\" " +
                "content=\"Robert.E.McGuire.1\" " + metaEnd);
            out.println("  <meta name=\"webmaster\" " +
                "content=\"Tamara.J.Kovalick.1\" " + metaEnd);
            out.println("  <link rel=\"icon\" href=\"nasaicon.png\" " +
                "type=\"image/png\" " + metaEnd);
            out.println(
                "  <link rel=\"SHORTCUT ICON\" href=\"favicon.ico\" " + 
                metaEnd);
            out.println("</head>");
            out.println("<body>");
        }
        else {

            out.println(
                "<!--#include virtual=\"../cdawheader_cgi.html\"-->");
        }
    }


    /**
     * Prints the table-of-contents portion of a dataset notes HTML 
     * page.
     *
     * @param datasets map containing the datasets that are to be 
     *            included in the table of contents
     * @param title title of HTML page
     * @param out PrintWriter to use for printing
     */
    private void PrintToc(SortedMap datasets, String title, 
                          PrintWriter out) {

        out.println("<h2>" + title + "</h2>");
        out.println("<dl>");

/*
        for (Iterator i = datasets.keySet().iterator(); i.hasNext(); ) {

            String dataset = (String)i.next();
                                       // a dataset name
            DatasetNotes dNotes = (DatasetNotes)datasets.get(dataset);
                                       // notes about dataset
*/
        for (Iterator i = datasets.entrySet().iterator(); i.hasNext(); ) {

            Map.Entry datasetEntry = (Map.Entry)i.next();
                                       // dataset entry
            String dataset = (String)datasetEntry.getKey();
                                       // a dataset name
            DatasetNotes dNotes = (DatasetNotes)datasetEntry.getValue();
                                       // notes about dataset
            String description = 
                TextUtils.htmlify(dNotes.getDescription());
                                       // XHTML version of description

            out.println("<dd><a href=\"#" + dataset + "\">" + dataset +
                        "</a>: " + description + "</dd>");
        }

        out.println("</dl>");
    }


    /**
     * Prints the notes portion of a dataset notes HTML page.
     *
     * @param datasets map containing the datasets that are to be 
     *            included in the page
     * @param out PrintWriter to use for printing
     */
    private void PrintNotes(SortedMap datasets, PrintWriter out) {

/*
        for (Iterator i = datasets.keySet().iterator(); i.hasNext(); ) {

            String dataset = (String)i.next();
                                       // a dataset name
            DatasetNotes dNotes = (DatasetNotes)datasets.get(dataset);
                                       // notes about dataset
*/
        for (Iterator i = datasets.entrySet().iterator(); i.hasNext(); ) {

            Map.Entry datasetEntry = (Map.Entry)i.next();
                                       // dataset entry
            String dataset = (String)datasetEntry.getKey();
                                       // a dataset name
            DatasetNotes dNotes = (DatasetNotes)datasetEntry.getValue();
                                       // notes about dataset

            out.println("<hr />");
            out.println("<dl>");
            out.println("<dt><a name=\"" + dataset + "\"></a>");
            out.println("<b>" + dataset + "</b></dt>");

            PrintGlobalNotes(dNotes, out);

            out.println("</dl>");
            out.println("<pre> </pre>");

            PrintVariableNotes(dNotes, out);

            out.println("<a href=\"/cgi-bin/eval2.cgi?dataset=" + dataset + "&index=sp_phys\">Dataset in CDAWeb</a>");
            out.println("<br/>");
            out.println("<a href=\"#top\">Back to top</a>");
        }
    }


    /**
     * Prints the global attribute notes portion of a dataset notes 
     * HTML page.
     *
     * @param dNotes map containing the datasets that are to be included
     *               in the page
     * @param out PrintWriter to use for printing
     */
    private void PrintGlobalNotes(DatasetNotes dNotes,
                                  PrintWriter out) {

        ArrayList value = null;        // global attribute value

        out.println("<dd><b>Description</b><pre>");

        value = dNotes.getText();

        if (value.size() > 0) {

            PrintText(value, out);
            PrintText(dNotes.getTextSupplement(), out);
            PrintText(dNotes.getTextSupplement1(), out);
        }
        else {

            out.println("No TEXT global attribute value.");
        }
        out.println("</pre></dd>");

        value = dNotes.getMods();

        if (value.size() > 0) {

            out.println("<dd><b>Modification History</b><pre>");
            PrintText(value, out);
            out.println("</pre></dd>");
        }

        value = dNotes.getCaveats();

        if (value.size() > 0) {

            out.println("<dd><b>Caveats</b><pre>");
            PrintText(value, out);
            out.println("</pre></dd>");
        }
    }


    /**
     * Prints the given array of text as fixed length lines in a 
     * representation that is suitable for a preformated portion of
     * an HTML document.  The text is broken on word boundaries and
     * transformed for use in an HTML document.  The transformations
     * performed include changing characters such as {@code '<'} to 
     * {@code "&lt;"} and changing URL strings to {@code <a href>} 
     * elements.
     *
     * @param text the text that is to be printed
     * @param out PrintWriter to use for printing
     */
    private void PrintText(ArrayList text, PrintWriter out) {

        for (int i = 0; i < text.size(); i++) {

            PrintText( (String)text.get(i), out);
        }
    }


    /**
     * Prints the given string of text as fixed length lines in a 
     * representation that is suitable for a preformated portion of
     * an HTML document.  The text is broken on word boundaries and
     * transformed for use in an HTML document.  The transformations
     * performed include changing characters such as {@code '<'} to 
     * {@code "&lt;"} and changing URL strings to {@code <a href>}
     * elements.
     *
     * @param text the text that is to be printed
     * @param out PrintWriter to use for printing
     */
    private void PrintText(String text, PrintWriter out) {

        if (text == null) {

            return;
        }

        String[] lines = TextUtils.splitBetweenWords(text, 80);
                                       // 80 character lines

        for (int i = 0; i < lines.length; i++) {

            out.println(TextUtils.hotlinkify(
                                    TextUtils.htmlify(lines[i]), "_TOP"));
        }
    }


    /**
     * Prints the variable attribute notes portion of a dataset notes 
     * HTML page.
     *
     * @param dNotes map containing the datasets that are to be included
     *               in the page
     * @param out PrintWriter to use for printing
     */
    private void PrintVariableNotes(DatasetNotes dNotes,
                                    PrintWriter out) {

        TreeMap varMap = dNotes.getVariableNotes();
                                       // variable notes map
//        Iterator varIter = varMap.keySet().iterator();
        Iterator varIter = varMap.entrySet().iterator();
                                       // variables iterator

        if (varIter.hasNext()) {

            out.println("<ul><li style=\"list-style: none\">" +
                        "<b>Variable Notes</b><pre></pre><dl>");
        }

        while (varIter.hasNext()) {

/*
            String varName = (String)varIter.next();
                                       // variable's name
            VariableNotes varNotes = (VariableNotes)varMap.get(varName);
                                       // variable's notes
*/
            Map.Entry varEntry = (Map.Entry)varIter.next();
                                       // variable map entry
            String varName = (String)varEntry.getKey();
                                       // variable's name
            VariableNotes varNotes = (VariableNotes)varEntry.getValue();
                                       // variable's notes

            PrintVariableNotes(varNotes, out);
        }

        if (varMap.size() > 0) {  // there were variables printed

            out.println("</dl></li></ul>");
        }
    }


    /**
     * Prints an individual variable's attribute notes portion of a 
     * dataset notes HTML page.
     *
     * @param varNotes the variable notes that are to be printed
     * @param out PrintWriter to use for printing
     */
    private void PrintVariableNotes(VariableNotes varNotes,
                                    PrintWriter out) {

        String name = TextUtils.htmlify(varNotes.getName());
                                       // XHTML version of variable name
        out.println("<dt><b>" + name + "</b></dt><dd><pre>");

        PrintText(varNotes.getNotes(), out);

        out.println("</pre></dd>");
    }


    /**
     * Prints the footer portion of an HTML page.
     *
     * @param out PrintWriter to use for printing
     */
    private void PrintFooter(PrintWriter out) {

        if (outputXhtml || outputHtml) {

            out.println("</body>");
            out.println("</html>");
        }
        else {
            out.println(
                "<!--#include virtual=\"../cdawfooter.html\"-->");
        }
    }


    /**
     * The main method used to invoke this application.
     *
     * @param args command line arguments
     * @throws IOException if an IOException occurs
     */
    public static void main(String[] args) 
        throws IOException {

        String outputFilename = "Notes";   
                                       // name of output file
        boolean noSubNotes = false;    // flag indicating that the sub
                                       // notes should not be created
        boolean outputXhtml = false;   // flag indicating that the 
                                       // output should be XHTML
        boolean outputHtml = false;    // output should be (standalone)
                                       // HTML
        int i = 0;                     // argv index
        for (; i < args.length && args[i].startsWith("-"); i++) {

            if (args[i].equals("-o")) {

                if (++i < args.length) {

                    outputFilename = args[i];
                }
                else {

                    exitArgsError("missing output-filename");
                }
            }
            else if (args[i].equals("-n")) {

                noSubNotes = true;
            }
            else if (args[i].equals("-x")) {

                outputXhtml = true;
            }
            else if (args[i].equals("-h")) {

                outputHtml = true;
            }
            else {

                exitArgsError("invalid option: " + args[i]);
            }
        }

        int numCdfs = 0;               // number of CDF files to process
        String[] cdfFiles = null;      // names of CDF files to process

        if (i == 0) {

            numCdfs = args.length;
            cdfFiles = args;
        }
        else { // args[i] was not an option or beyond end

            numCdfs = args.length - i;

            cdfFiles = new String[numCdfs];
            System.arraycopy(args, i, cdfFiles, 0, numCdfs);
        }

        if (numCdfs == 0) {

            exitArgsError("no CDF files given");
        }

        GetNotes getNotes = new GetNotes();

        getNotes.setOutputXhtml(outputXhtml);
        getNotes.setOutputHtml(outputHtml);
        getNotes.readNotes(cdfFiles);

        getNotes.writeAllNotes(outputFilename);

        if (!noSubNotes) {

            getNotes.writeAllSubNotes(outputFilename);
        }
    }


    /**
     * Terminates this application after printing the given message about
     * an "args" error and a message describing the application's
     * command line argument syntax.
     *
     * @param message text of a message describing a problem with the
     *                command-line arguments given to this application
     */
    public static void exitArgsError(String message) {

        System.err.println("Error: " + message);
        System.err.println(
            "Usage: GetNotes [-o output-filename] [-n] [-x] CDF-files");
        System.err.println("Where:");
        System.err.println(
            "    output-filename is the name of the file with the notes");
        System.err.println(
            "    -n specifies not to create the sub-notes files");
        System.err.println(
            "    -x specifies that the output is to be XHTML instead " +
            "of HTML");
        System.err.println(
            "    -h specifies that the output is to be standalone " +
            "HTML instead of HTML that depends upon include files");
        System.err.println(
            "    CDF-files are the names of the CDF files to extract the");
        System.err.println(
            "    notes from.  If multiple files are given for the same");
        System.err.println(
            "    dataset, the non-null, non-blank attribute values from");
        System.err.println(
            "    the latest file given will have preference.");
        System.exit(1);
    }
}
